"""Exports for the Pypechain package."""

# This tool is meant to be used as a command line tool.  It is exported here for testing.  Do not
# rely on this being available in the future.
from .main import main as pypechain_cli



"""Functions to render Python files from an abi usng a jinja2 template."""

from pathlib import Path

from jinja2 import Template

from pypechain.utilities.abi import (
    get_abi_items,
    get_input_names,
    get_input_names_and_values,
    get_output_names,
    is_abi_constructor,
    is_abi_function,
    load_abi_from_file,
)
from pypechain.utilities.format import capitalize_first_letter_only
from pypechain.utilities.sort import get_intersection_and_unique


def render_contract_file(
    contract_name: str, contract_template: Template, abi_file_path: Path
) -> str:
    """Returns the serialized code of the contract file to be generated.

    Arguments
    ---------
    contract_template : Template
        A jinja template containging types for all structs within an abi.
    abi_file_path : Path
        The path to the abi file to parse.

    Returns
    -------
    str
        A serialized python file.
    """

    # TODO:  return types to function calls
    # Extract function names and their input parameters from the ABI
    function_datas = {}
    constructor_data = {}
    for abi_function in get_abi_items(abi_file_path):
        if is_abi_function(abi_function):
            # TODO: investigate better typing here?  templete.render expects an object so we'll have
            # to convert.

            # hanndle constructor
            if is_abi_constructor(abi_function):
                constructor_data = {
                    "input_names_and_types": [get_input_names_and_values(abi_function)],
                    "input_names": [get_input_names(abi_function)],
                    "outputs": [get_output_names(abi_function)],
                }

            # handle functions
            else:
                name = abi_function.get("name", "")
                if name and name not in function_datas:
                    function_data = {
                        # TODO: pass a typeguarded ABIFunction that has only required fields?
                        # name is required in the typeguard.  Should be safe to default to empty string.
                        "name": name,
                        "capitalized_name": capitalize_first_letter_only(name),
                        "input_names_and_types": [
                            get_input_names_and_values(abi_function)
                        ],
                        "input_names": [get_input_names(abi_function)],
                        "outputs": [get_output_names(abi_function)],
                    }
                    function_datas[name] = function_data
                else:  # this function already exists, presumably with a different signature
                    function_datas[name]["input_names_and_types"].append(
                        get_input_names_and_values(abi_function)
                    )
                    function_datas[name]["input_names"].append(
                        get_input_names(abi_function)
                    )
                    function_datas[name]["outputs"].append(
                        get_output_names(abi_function)
                    )
                    # input_names_and_types will need optional args at the end
                    (
                        shared_input_names_and_types,
                        unique_input_names_and_types,
                    ) = get_intersection_and_unique(
                        function_datas[name]["input_names_and_types"]
                    )
                    function_datas[name][
                        "required_input_names_and_types"
                    ] = shared_input_names_and_types
                    function_datas[name]["optional_input_names_and_types"] = []
                    for name_and_type in unique_input_names_and_types:  # optional args
                        name_and_type += " | None = None"
                        function_datas[name]["optional_input_names_and_types"].append(
                            name_and_type
                        )
                    # we will also need the names to be separated
                    (
                        shared_input_names,
                        unique_input_names,
                    ) = get_intersection_and_unique(function_datas[name]["input_names"])
                    function_datas[name]["required_input_names"] = shared_input_names

                    function_datas[name]["optional_input_names"] = unique_input_names

    abi = load_abi_from_file(abi_file_path)

    # Render the template
    return contract_template.render(
        abi=abi,
        contract_name=contract_name,
        functions=list(function_datas.values()),
        # TODO: use this data to add a typed constructor
        constructor=constructor_data,

    )



"""Functions to render Python types from an abi usng a jinja2 template."""
from dataclasses import asdict
from pathlib import Path

from jinja2 import Template

from pypechain.utilities.abi import (
    get_events_for_abi,
    get_structs_for_abi,
    load_abi_from_file,
)


def render_types_file(
    contract_name: str, types_template: Template, abi_file_path: Path
) -> str:
    """Returns the serialized code of the types file to be generated.

    Arguments
    ---------
    contract_name : str
        The name of the contract to be parsed.
    types_template : Template
        A jinja template containging types for all structs within an abi.
    abi_file_path : Path
        The path to the abi file to parse.

    Returns
    -------
    str
        A serialized python file.
    """

    abi = load_abi_from_file(abi_file_path)

    structs_by_name = get_structs_for_abi(abi)
    structs_list = list(structs_by_name.values())
    structs = [asdict(struct) for struct in structs_list]
    events = [asdict(event) for event in get_events_for_abi(abi)]
    has_events = bool(events)
    has_event_params = any(len(event["inputs"]) > 0 for event in events)

    return types_template.render(
        contract_name=contract_name,
        structs=structs,
        events=events,
        has_events=has_events,
        has_event_params=has_event_params,
    )



"""Functions to render Python files from an abi usng a jinja2 template."""

import os
from pathlib import Path

from pypechain.render.contract import render_contract_file
from pypechain.render.types import render_types_file
from pypechain.utilities.file import write_string_to_file
from pypechain.utilities.format import apply_black_formatting
from pypechain.utilities.templates import setup_templates


def render_files(abi_file_path: str, output_dir: str, line_length: int) -> None:
    """Processes a single JSON file to generate class and types files."""

    # get names
    file_path = Path(abi_file_path)
    output_path = Path(output_dir)
    filename = file_path.name
    contract_name = os.path.splitext(filename)[0]
    contract_path = output_path.joinpath(f"{contract_name}")

    # grab the templates
    contract_template, types_template = setup_templates()

    # render the code
    rendered_contract_code = render_contract_file(
        contract_name, contract_template, file_path
    )
    rendered_types_code = render_types_file(contract_name, types_template, file_path)

    # Format the generated code using Black
    formatted_contract_code = apply_black_formatting(
        rendered_contract_code, line_length
    )
    formatted_types_code = apply_black_formatting(rendered_types_code, line_length)

    # Write the code to file
    write_string_to_file(f"{contract_path}Contract.py", formatted_contract_code)
    write_string_to_file(f"{contract_path}Types.py", formatted_types_code)
